import numpy as np
from hw2_utils import TestCase, run_tests


class TestNumpy(TestCase):
    def __init__(self, *args, **kwargs):
        self.module_name = "hw2_numpy"
        TestCase.__init__(self, *args, **kwargs)

    def test01_convolution(self):
        """Test the implementation of Convolutional layer."""

        np.random.seed(4096)

        # convolution spec:
        # - x: Input data of shape (N, C, H, W)
        # - w: Filter weights of shape (F, C, HH, WW)

        N = 4
        C = 3
        H = 224
        W = 224
        F = 3
        HH = 3
        WW = 3

        FILTER0 = 0
        FILTER1 = 1
        FILTER2 = 2

        RED = 0
        GREEN = 1
        BLUE = 2

        x = np.random.randn(N, C, H, W)

        # Set up a convolutional weights holding 3 filters of 3x3 pixels over 3
        # channels
        w = np.zeros((F, C, HH, WW))

        # Filter over the red channel.
        w[FILTER0, RED, :, :] = [[1, 2, 1],
                                 [0, 0, 0],
                                 [-1, -2, -1]]

        # Luma coding filter.
        w = self.solution.grayscale_filter(w)

        # Vertical edge detector over the green channel.
        w[FILTER2, GREEN, :, 1] = [-0.5, 1.0, -0.5]

        # Vector of biases, shape = (F)
        b = np.array([0, 42, 37])

        # Compute the result of convolving each input in x with each filter in
        # w, offsetting/biasing by b, and storing the results in out.
        out = self.solution.convolution(
            x, w, b,
            stride=1, pad=1
        )

        self.assertIsInstance(out, np.ndarray,
                              msg="Conv output should be a numpy array.")

        self.assertEqual(hash(out.shape) % 0xffffffff,
                         5190712560950120328 % 0xffffffff,
                         msg="Conv output shape incorrect.")

        # a 2x2 region in output of all three filters on instance 0
        actual = out[0, :, 2:4, 6:8]
        expected = np.array(
            [[[2.36845895, -0.38982775],
              [1.37878635,  1.4613423 ]],
             [[42.71596034, 42.11010472],
              [40.99883455, 42.56813345]],
             [[38.54311008, 35.58869492],
              [35.69644451, 38.16163285]]])

        self.assertArrayEqual(actual,
                              expected,
                              msg='Output of conv layer is wrong.')

    def test02_relu(self):
        """Test the implementation of relu layer."""

        np.random.seed(4)
        x = np.random.randn(5, 5, 32)
        out = self.solution.relu(x)

        self.assertIsInstance(out, np.ndarray,
                              msg="RELU output should be a numpy array.")

        self.assertShapeEqual(
            out.shape, x.shape,
            msg="RELU output shape must match its input shape."
        )

        actual = x[4, 2, 12:].sum() - out[4, 2, 12:].sum()
        expected = -3.3328890425375954

        self.assertAlmostEqual(
            actual, expected,
            msg=('Your relu implementation is wrong. Make sure you do not '
                 'in-place modify the input.'),
            delta=1e-14)

    def test_maxpool(self):
        """Test the implementation of Maxpool layer."""

        np.random.seed(67)
        x = np.random.randn(4, 3, 24, 42)
        pool_param = {'pool_height': 2, 'pool_width': 3, 'stride': 2}
        out = self.solution.max_pool(x, pool_param)

        self.assertIsInstance(out, np.ndarray,
                              msg="Maxpool output should be a numpy array.")

        actual_shape = out.shape
        expected_shape = (4, 3, 12, 20)

        self.assertShapeEqual(
            actual_shape,
            expected_shape,
            msg='Output shape does not match the expected shape.')

    def test_dropout(self):
        """Test the implementation of dropout layer"""

        np.random.seed(7)
        x = np.abs(np.random.randn(1000, 3, 24, 42))
        size = x.size

        sum = x.sum()

        result_train = self.solution.dropout(x.copy(), 'train', 0.2)
        result_test = self.solution.dropout(x.copy(), 'test', 0.4)

        self.assertIsInstance(
            result_train, np.ndarray,
            msg="Dropout train mode output should be a numpy array."
        )
        self.assertIsInstance(
            result_test, np.ndarray,
            msg="Dropout test mode output should be a numpy array."
        )

        self.assertShapeEqual(
            result_train.shape, x.shape,
            msg="Dropout output shape must be same as input shape."
        )

        self.assertShapeEqual(
            result_test.shape, x.shape,
            msg="Dropout output shape must be same as input shape."
        )

        dropped_train = (result_train > 0).sum()
        dropped_test = (result_test > 0).sum()

        self.assertAlmostEqual(
            dropped_train / size,
            0.8,
            places=2,
            msg="Dropped the wrong amount of activations in train mode.")

        self.assertAlmostEqual(
            dropped_test / size,
            1.0,
            places=2,
            msg="Dropped the wrong amount of activations in test mode.")

        self.assertAlmostEqual(
            result_train.sum() / sum,
            1.0,
            places=2,
            msg="Deviation in expected activations in train mode.")

        self.assertAlmostEqual(
            result_test.sum() / sum,
            1.0,
            places=2,
            msg="Deviation in expected activations in test mode.")


if __name__ == '__main__':
    run_tests(TestNumpy)
